from unittest import mock
from os import environ
from dotenv import load_dotenv

import pytest
from osdatahub.NGD.crs import get_crs
from osdatahub.NGD.ngd_api import _merge_geojsons, NGD
from osdatahub.NGD.models import NGDFeatureCollection

from tests.data import ngd_crs_data as crs_data
from tests.data import ngd_merge_geojsons_data as merge_geojsons_data
from tests.data import ngd_query_data as query_data

load_dotenv()

API_KEY = environ.get("OSDATAHUB_TEST_KEY")


class TestNGDCRS:
    @pytest.mark.parametrize(*crs_data.test_ngd_crs())
    def test_ngd_crs_pass(self, crs, valid_crs, expected_result):
        assert get_crs(crs, valid_crs) == expected_result

    @pytest.mark.parametrize(*crs_data.test_ngd_crs_fail())
    def test_ngd_crs_fail(self, crs, valid_crs, expected_result):
        with pytest.raises(expected_exception=expected_result):
            get_crs(crs, valid_crs)


class TestNGDQuery:
    @pytest.mark.parametrize(*query_data.test_ngd_instantiation())
    @mock.patch('osdatahub.get')
    def test_ngd_instantiation(self, request_mocked, api_key, collection, expected_url, expected_headers):
        request_mocked.return_value.configure_mock(json=lambda: {"features": [], "numberReturned": 0})

        ngd = NGD(key=api_key, collection=collection)
        ngd.query()

        request_mocked.assert_called_with(expected_url,
                                          headers=expected_headers,
                                          params={"limit": 100, "offset": 0},
                                          proxies={})

    @pytest.mark.parametrize(*query_data.test_ngd_query())
    @mock.patch('osdatahub.get')
    def test_ngd_api_call(self, request_mocked, extent, crs, start_datetime, end_datetime, cql_filter, filter_crs,
                          max_results, offset, expected_url, expected_params):
        request_mocked.return_value.configure_mock(json=lambda: {"features": [], "numberReturned": 0})

        ngd = NGD(key="API-KEY", collection="bld-fts-buildingline")

        ngd.query(extent=extent,
                  crs=crs,
                  start_datetime=start_datetime,
                  end_datetime=end_datetime,
                  cql_filter=cql_filter,
                  filter_crs=filter_crs,
                  max_results=max_results,
                  offset=offset
                  )
        request_mocked.assert_called_with(expected_url,
                                          params=expected_params,
                                          headers={"key": "API-KEY"},
                                          proxies={})

    @pytest.mark.parametrize(*query_data.test_ngd_query_fail())
    def test_ngd_api_call_fail(self, extent, crs, start_datetime, end_datetime, cql_filter, filter_crs,
                               max_results, offset, expected_error):
        ngd = NGD("API_KEY", "blt-fts-buildingline")
        with pytest.raises(expected_exception=expected_error):
            ngd.query(extent=extent,
                      crs=crs,
                      start_datetime=start_datetime,
                      end_datetime=end_datetime,
                      cql_filter=cql_filter,
                      filter_crs=filter_crs,
                      max_results=max_results,
                      offset=offset
                      )

    @pytest.mark.skipif(not API_KEY, reason="Test API key not available")
    @pytest.mark.parametrize(*query_data.test_ngd_query_live())
    def test_ngd_api_call(self, collection, extent, crs, start_datetime, end_datetime, cql_filter, filter_crs,
                          max_results, offset):
        ngd = NGD(key=API_KEY, collection=collection)
        results = ngd.query(extent=extent,
                  crs=crs,
                  start_datetime=start_datetime,
                  end_datetime=end_datetime,
                  cql_filter=cql_filter,
                  filter_crs=filter_crs,
                  max_results=max_results,
                  offset=offset
                )

        assert len(results["features"]) == max_results

    @pytest.mark.skipif(not API_KEY, reason="Test API key not available")
    @pytest.mark.parametrize(*query_data.test_ngd_query_live())
    def test_ngd_api_call_collection(self, collection, extent, crs, start_datetime, end_datetime, cql_filter, filter_crs,
                          max_results, offset):
        ngd = NGD(key=API_KEY, collection=collection)
        results = ngd.query(extent=extent,
                  crs=crs,
                  start_datetime=start_datetime,
                  end_datetime=end_datetime,
                  cql_filter=cql_filter,
                  filter_crs=filter_crs,
                  max_results=max_results,
                  offset=offset,
                  output_as_collection=True
                )

        assert type(results) is NGDFeatureCollection
        assert len(results.features) == max_results

class TestNGDGetCollections:
    @mock.patch('osdatahub.get')
    def test_ngd_get_collections(self, request_mocked):
        request_mocked.return_value.configure_mock(json=lambda: {})
        NGD.get_collections()

        request_mocked.assert_called_with("https://api.os.uk/features/ngd/ofa/v1/collections",
                                          proxies={})


class TestNGDQueryFeature:

    @pytest.mark.parametrize(*query_data.test_ngd_query_feature())
    @mock.patch('osdatahub.get')
    def test_ngd_query_filter(self, request_mocked, feature_id, crs, expected_url, expected_params):
        request_mocked.return_value.configure_mock(json=lambda: {})
        ngd = NGD("api_key", "bld-fts-buildingline")
        ngd.query_feature(feature_id=feature_id, crs=crs)

        request_mocked.assert_called_with(expected_url,
                                          params=expected_params,
                                          headers={"key": "api_key"},
                                          proxies={})


class TestMergeGeojsons:
    @pytest.mark.parametrize(*merge_geojsons_data.test_merge_geojsons_pass())
    def test_merge_geojsons_pass(self, gj1, gj2, expected_result):
        m = _merge_geojsons(gj1, gj2)
        assert m["numberReturned"] == expected_result
        assert len(m["features"]) == expected_result

    @pytest.mark.parametrize(*merge_geojsons_data.test_merge_geojsons_fail())
    def test_merge_geojsons_fail(self, gj1, gj2, expected_result):
        with pytest.raises(expected_exception=expected_result):
            _merge_geojsons(gj1, gj2)



import json
import logging
import os

os.environ["_OSDATAHUB_PROXIES"] = json.dumps({})


def set_proxies(proxies):
    os.environ["_OSDATAHUB_PROXIES"] = json.dumps(proxies)


def get_proxies():
    return json.loads(os.environ["_OSDATAHUB_PROXIES"])

__version__ = "1.3.3"

from osdatahub.DownloadsAPI import DataPackageDownload, OpenDataDownload
from osdatahub.extent import Extent
from osdatahub.FeaturesAPI import FeaturesAPI
from osdatahub.LinkedIdentifiersAPI import LinkedIdentifiersAPI
from osdatahub.NamesAPI import NamesAPI
from osdatahub.NGD import NGD, AsyncNGD
from osdatahub.PlacesAPI import PlacesAPI
from osdatahub.requests_wrapper import get, post



import json
from os import environ

import geojson
import requests
from geojson import FeatureCollection
from typeguard import typechecked

import osdatahub
from osdatahub import Extent, FeaturesAPI
from osdatahub.errors import raise_http_error
from osdatahub.filters import is_equal
from osdatahub.grow_list import GrowList

"""
script to generate unprocessed API data
"""


class FeaturesAPI_NoPostProcessing(FeaturesAPI):
    @typechecked
    def query(self, limit: int = 1000) -> FeatureCollection:
        """Run a query of the OS Features API WITHOUT POST PROCESSING

        Args:
            limit (int, optional): The maximum number of features to return.
                Defaults to 1000.

        Returns:
            FeatureCollection: The results of the query in GeoJSON format
        """
        params = self._FeaturesAPI__params()
        data = GrowList()
        n_required = min(limit, 100)
        try:
            while n_required > 0 and data.grown:
                params.update({"count": n_required, "startIndex": len(data)})
                response = osdatahub.get(self.ENDPOINT, params=params, proxies=osdatahub.get_proxies())
                data.extend(response.json()["features"])
                n_required = min(100, limit - len(data))
        except json.decoder.JSONDecodeError:
            raise_http_error(response)
        # DOESN'T run features_to_geojson before returning result
        return FeatureCollection(data.values, crs=self.extent.crs)


def round_recursively(lst):
    if isinstance(lst, (float, int)):
        return round(lst, 2)
    else:
        return [round_recursively(i) for i in lst]


if __name__ == "__main__":

    # Need equivelant files via QGIS. Export features WITHOUT any properties.
    test_polygons = [
        (
            "osgb4000000073320773",
            (391436, 172360, 393874, 174833),
            "sites_multipolygon",
        ),
        (
            "osgb4000000073320772",
            (391436, 172360, 393874, 174833),
            "sites_polygon_with_hole",
        ),
        ("osgb4000000073320770", (391436, 172360, 393874, 174833), "sites_polygon"),
        (
            "osgb4000000073320769",
            (391436, 172360, 393874, 174833),
            "sites_multipolygon_with_hole",
        ),
        (
            "osgb4000000073495159",
            (336610, 503920, 338001, 505331),
            "sites_multipolygon2",
        ),
    ]

    key = environ.get("OS_API_KEY")
    folder = r"tests\data\clean_polygon_data\API (uncleaned)"
    for TOID, bbox, save_name in test_polygons:
        extent = Extent.from_bbox(bbox, "EPSG:27700")
        features = FeaturesAPI_NoPostProcessing(key, "sites_functional_site", extent)
        features.add_filters(is_equal("TOID", TOID))
        results = features.query(limit=42)

        for site in results["features"]:
            coords = site["geometry"]["coordinates"]
            site["geometry"]["coordinates"] = round_recursively(coords)
            site["properties"] = {}
        with open(rf"{folder}\{save_name}.geojson", "w", encoding="utf-8") as f:
            geojson.dump(
                FeatureCollection(results, crs="EPSG:27700"),
                f,
                ensure_ascii=False,
                indent=4,
            )
            print(f"{save_name} saved")



import functools
import json
import logging
from datetime import datetime
from typing import Union

import requests
from typeguard import typechecked

import osdatahub
from osdatahub import Extent
from osdatahub.NGD.crs import get_crs
from osdatahub.NGD.models import NGDFeatureCollection


def _merge_geojsons(gj1: Union[dict], gj2: Union[dict]) -> Union[dict]:
    """
    Combines 2 geojsons from NGD api into a single valid geojson

    Args:
        gj1 (FeatureCollection): A FeatureCollection
        gj2 (FeatureCollection): Another FeatureCollection

    Returns (FeatureCollection): A FeatureCollection with a single set of Features and a combined numberReturned

    """
    if not (gj1 or gj2):
        raise ValueError("Inputs were both empty")
    elif not gj1:
        return gj2
    elif not gj2:
        return gj1

    required_keys = {"features", "numberReturned", "links"}
    if not (gj1.keys() >= required_keys and gj2.keys() >= required_keys):
        raise ValueError(f"Both geojsons must contain keys {required_keys}")

    merged_geojson = gj1.copy()
    merged_geojson["features"] = merged_geojson["features"] + gj2["features"]
    merged_geojson["numberReturned"] += gj2["numberReturned"]
    merged_geojson["links"] += gj2["links"]

    return merged_geojson


class NGD:
    """
    Main class for querying OS NGD Features API (https://osdatahub.os.uk/docs/ofa/overview)

    Args:
        key (str): A valid OS Data Hub API key. Get a free key here - https://osdatahub.os.uk/
        collection (str): ID for the desired NGD Feature Collection. Learn about the possible collection ids here -
            https://osdatahub.os.uk/docs/ofa/technicalSpecification

    Example::

        from osdatahub import NGD
        from os import environ

        key = environ.get("OS_API_KEY")
        extent = Extent.from_bbox((600000, 310200, 600900, 310900), "EPSG:27700")
        features = NGD(key, "bld-fts-buildingline")
        results = features.query(max_results=50, extent=extent)
    """

    __ENDPOINT = r"https://api.os.uk/features/ngd/ofa/v1/collections"

    __HEADERS = {"Accept": "application/geo+json"}

    def __init__(self, key: str, collection: str):
        self.key: str = key
        self.collection: str = collection

    def __endpoint(self, feature_id=None) -> str:
        return f"{self.__ENDPOINT}/{self.collection}/items/{feature_id if feature_id else ''}"

    @classmethod
    @functools.lru_cache()
    def get_collections(cls) -> dict:
        """
        Retrieves all OS NGD Feature Collections
        Returns:
            Dict: Dictionary containing all Feature Collections currently supported with details for each
        """
        response = osdatahub.get(cls.__ENDPOINT, proxies=osdatahub.get_proxies())
        response.raise_for_status()
        return response.json()

    @typechecked
    def query(
        self,
        extent: Union[Extent, None] = None,
        crs: Union[str, int, None] = None,
        start_datetime: Union[datetime, None] = None,
        end_datetime: Union[datetime, None] = None,
        cql_filter: Union[str, None] = None,
        filter_crs: Union[str, int, None] = None,
        max_results: int = 100,
        offset: int = 0,
        output_as_collection: bool = False
    ) -> Union[dict, NGDFeatureCollection]:
        """
        Retrieves features from a Collection

        Args:
            extent (Extent, optional): An extent, either from a bounding box or a Polygon. Only features within this
                extent will be returned
                Available CRS values are: EPSG:27700, EPSG:4326, EPSG:3857, and CRS84. Defaults to CRS84
            crs (str|int, optional): The CRS for the returned features, either in the format "epsg:xxxx" or an epsg
                number. e.g. British National Grid can be supplied as "epsg:27700" or 27700.
                Available CRS values are: EPSG:27700, EPSG:4326, EPSG:7405, EPSG:3857, and CRS84. Defaults to CRS84
            start_datetime (datetime, optional): Selects features that have a temporal property after the given start
                time. If you want to query a single timestamp, provide the same value to both start_datetime
                and end_datetime
            end_datetime (datetime, optional): Selects features that have a temporal property before the given end
                time. If you want to query a single timestamp, provide the same value to both start_datetime
                and end_datetime
            cql_filter (str, optional): A filter query in the CQL format. More information about supported CQL operators
                can be found at https://osdatahub.os.uk/docs/ofa/technicalSpecification
            filter_crs (str|int, optional): The CRS for a given CQL query, either in the format "epsg:xxxx" or an epsg
                number. e.g. British National Grid can be supplied as "epsg:27700" or 27700
                Available CRS values are: EPSG:27700, EPSG:4326, EPSG:7405, EPSG:3857, and CRS84. Defaults to CRS84
            max_results (int, optional): The maximum number of features to return. Defaults to 100
            offset (int, optional): The offset number skips past the specified number of features in the collection.
                Defaults to 0
            output_as_collection (bool, optional): Outputs as a NGDFeatureCollection which is more pythonic and structured.
                Defaults to False to ensure backwards compatibility. 
        Returns either:
            Dict: The results of the query in GeoJSON format (if output_as_collection is set to False)
            NGDFeatureCollection: The results of the query in structured form (if output_as_collection is set to True)
        """

        assert max_results > 0, (
            f"Argument max_results must be greater than 0 but was {max_results}"
        )
        assert offset >= 0, f"Argument offset must be greater than 0 but was {offset}"
        params = {}

        # Checking validity and preformatting arguments
        if crs:
            params["crs"] = get_crs(crs=crs)

        if start_datetime or end_datetime:
            if start_datetime and end_datetime and start_datetime > end_datetime:
                raise ValueError("Start time must be before end time")

            start_datetime = (
                start_datetime.isoformat() + "Z" if start_datetime else ".."
            )
            end_datetime = end_datetime.isoformat() + "Z" if end_datetime else ".."

            params["datetime"] = f"{start_datetime}/{end_datetime}"

        if extent:
            bbox_filter = f"INTERSECTS(geometry, {extent.polygon.wkt})"

            # ADD INTERSECTS QUERY
            if cql_filter:
                if filter_crs:
                    assert get_crs(
                        extent.crs,
                        valid_crs=("epsg:4326", "epsg:27700", "epsg:3857", "crs84"),
                    ) == get_crs(filter_crs), (
                        "If passing extent and a cql filter with a crs, the filter_crs must "
                        "be same as the extent crs"
                    )
                else:
                    filter_crs = extent.crs

                cql_filter += f" AND {bbox_filter}"
            else:
                cql_filter = bbox_filter
                filter_crs = extent.crs

        if cql_filter:
            params["filter"] = cql_filter
            if filter_crs:
                params["filter-crs"] = get_crs(
                    crs=filter_crs,
                    valid_crs=("epsg:4326", "epsg:27700", "epsg:3857", "crs84"),
                )

        n_required = max_results

        data = {}

        headers = self.__HEADERS
        headers.update({"key": self.key})

        while n_required > 0:
            limit = min(n_required, 100)
            offset = max(
                offset, data["numberReturned"] if "numberReturned" in data else 0
            )
            params.update({"limit": limit, "offset": offset})
            try:
                response = osdatahub.get(
                    self.__endpoint(),
                    params=params,
                    headers=headers,
                    proxies=osdatahub.get_proxies(),
                )
                response.raise_for_status()
            except requests.exceptions.HTTPError as e:
                logging.error(json.dumps(e.response.json(), indent=4))
                raise e

            resp_json = response.json()
            data = _merge_geojsons(data, resp_json)

            if resp_json["numberReturned"] < limit:
                break
            else:
                n_required -= resp_json["numberReturned"]

        # Added to allow the NGD Sync to work with the NGD ASync
        if output_as_collection:
            data = NGDFeatureCollection.from_dict(data)

        return data

    def query_feature(self, 
                      feature_id: str, 
                      crs: Union[str, int] = None
                      ) -> dict:
        """
        Retrieves a single feature from a collection

        Args:
            feature_id: An identifier ID for a feature
            crs (str|int, optional): The CRS for the returned feature, either in the format "epsg:xxxx" or an epsg
                number. e.g. British National Grid can be supplied as "epsg:27700" or 27700.
                Available CRS values are: EPSG:27700, EPSG:4326, EPSG:7405, EPSG:3857, and CRS84. Defaults to CRS84

        Returns:
            Feature: A GeoJSON Feature containing the requested feature
        """
        params = {"crs": get_crs(crs)} if crs else {}

        response = osdatahub.get(
            self.__endpoint(feature_id),
            params=params,
            headers={"key": self.key},
            proxies=osdatahub.get_proxies(),
        )
        response.raise_for_status()

        return response.json()



import json

import osdatahub
import requests

ID_ENDPOINT = "http://statistics.data.gov.uk/geometry?resource=http://statistics.data.gov.uk/id/statistical-geography/"


def _sanitise_response(res: requests.Response) -> dict:
    res_body = res.text
    res_json = json.loads(res_body)
    res_json = _remove_duplicate_features(res_json)
    return res_json["features"][0]["geometry"]


def _remove_duplicate_features(response_json: dict) -> dict:
    """For some reason the API json response contains either 1 geomety or
        duplicates the same entry twice

    Args:
        response_json (dict): The json of the raw response from the ONS API

    Returns:
        response_json (dict): The json of the raw response with duplicates removed

    Raises:
        ValueError: If the raw response doesn't contain 1 or 2 features
    """
    if len(response_json["features"]) == 1:
        return response_json
    elif len(response_json["features"]) == 2:
        response_json["features"].pop(1)
        return response_json
    else:
        raise ValueError(f"Response JSON should contains either 1 or 2.")


def get_ons_geom(ons_code: str) -> dict:
    """Gets coordinates of the boundary of the polygon for the ons_code

    Args:
        ons_code (str): ONS code

    Returns:
        response_json (dict): The json of the raw response
    """
    url = f"{ID_ENDPOINT}{ons_code}"
    res = osdatahub.get(url, proxies=osdatahub.get_proxies())
    return _sanitise_response(res)



from osdatahub.extent import Extent
from pytest import param
from typeguard import TypeCheckError


def test_format_fq():
    test_variables = "bbox, local_type, expected_result"
    test_data = [
        param(None,
              "oil_refining",
              ["LOCAL_TYPE:Oil_Refining"]),
        param(None,
              "special_needs_education",
              ["LOCAL_TYPE:Special_Needs_Education"]),
        param(None,
              ["town", "road_user_services"],
              ["LOCAL_TYPE:Town LOCAL_TYPE:Road_User_Services"]),
        param(None,
              ["hill_or_mountain", "heliport", "postcode",
                  "bay", "electricity_distribution"],
              ["LOCAL_TYPE:Hill_Or_Mountain LOCAL_TYPE:Heliport LOCAL_TYPE:Postcode LOCAL_TYPE:Bay "
               "LOCAL_TYPE:Electricity_Distribution"]),
        param(Extent.from_bbox((1000, 1000, 2000, 2000), crs="EPSG:27700"),
              None,
              ["BBOX:1000.0,1000.0,2000.0,2000.0"]),
        param(Extent.from_bbox((10000, 20000, 30000, 40000), crs="EPSG:27700"),
              None,
              ["BBOX:10000.0,20000.0,30000.0,40000.0"]),
        param(Extent.from_bbox((10000, 20000, 30000, 40000), crs="EPSG:27700"),
              "estuary",
              ["LOCAL_TYPE:Estuary", "BBOX:10000.0,20000.0,30000.0,40000.0"]),
        param(Extent.from_bbox((1000, 1000, 2000, 2000), crs="EPSG:27700"),
              ["other_settlement", "passenger_ferry_terminal", "spot_height"],
              ["LOCAL_TYPE:Other_Settlement LOCAL_TYPE:Passenger_Ferry_Terminal LOCAL_TYPE:Spot_Height",
               "BBOX:1000.0,1000.0,2000.0,2000.0"]),

    ]
    return test_variables, test_data


def test_format_fq_errors():
    test_variables = "bbox, local_type, expected_result"
    test_data = [
        param(None,
              1234,
              (TypeError,TypeCheckError)),
        param(1234,
              None,
              (TypeError,TypeCheckError)),
        param(None,
              "fake_local_type",
              ValueError),
        param(None,
              ["town", "estuary", "fake_local_type"],
              ValueError),
        param(Extent.from_bbox((1000, 2000, 3000, 4000), crs="EPSG:4326"),
              None,
              ValueError)
    ]
    return test_variables, test_data


def test_find_pass():
    test_variables = "text, limit, bounds, bbox_filter, local_type, expected_url, expected_params"
    test_data = [
        param("Buckingham Palace",
              100,
              None,
              None,
              None,
              'https://api.os.uk/search/names/v1/find?key=test',
              {"query": "Buckingham Palace", "offset": 0, "maxresults": 100}
              ),
        param("OS HQ",
              100,
              None,
              None,
              None,
              'https://api.os.uk/search/names/v1/find?key=test',
              {"query": "OS HQ", "offset": 0, "maxresults": 100}
              ),
        param("Buckingham Palace",
              50,
              None,
              None,
              None,
              'https://api.os.uk/search/names/v1/find?key=test',
              {"query": "Buckingham Palace", "offset": 0, "maxresults": 50}
              ),
        param("Buckingham Palace",
              50,
              Extent.from_bbox((100, 200, 300, 400), crs="EPSG:27700"),
              None,
              None,
              'https://api.os.uk/search/names/v1/find?key=test',
              {"query": "Buckingham Palace", "offset": 0,
                  "bounds": "100.0,200.0,300.0,400.0", "maxresults": 50}
              ),
        param("Buckingham Palace",
              50,
              None,
              Extent.from_bbox((100, 200, 300, 400), crs="EPSG:27700"),
              None,
              'https://api.os.uk/search/names/v1/find?key=test',
              {"query": "Buckingham Palace", "offset": 0, "fq": [
                  "BBOX:100.0,200.0,300.0,400.0"], "maxresults": 50}
              ),
        param("Buckingham Palace",
              50,
              None,
              None,
              "suburban_area",
              'https://api.os.uk/search/names/v1/find?key=test',
              {"query": "Buckingham Palace", "offset": 0, "fq": [
                  "LOCAL_TYPE:Suburban_Area"], "maxresults": 50}
              ),
        param("Buckingham Palace",
              50,
              None,
              None,
              ("suburban_area", "tramway"),
              'https://api.os.uk/search/names/v1/find?key=test',
              {"query": "Buckingham Palace", "offset": 0, "fq": ["LOCAL_TYPE:Suburban_Area LOCAL_TYPE:Tramway"],
               "maxresults": 50}
              ),
        param("Buckingham Palace",
              50,
              Extent.from_bbox((100, 200, 300, 400), crs="EPSG:27700"),
              None,
              "suburban_area",
              'https://api.os.uk/search/names/v1/find?key=test',
              {"query": "Buckingham Palace", "offset": 0, "bounds": "100.0,200.0,300.0,400.0",
               "fq": ["LOCAL_TYPE:Suburban_Area"], "maxresults": 50}
              ),
        param("Buckingham Palace",
              50,
              Extent.from_bbox((100, 200, 300, 400), crs="EPSG:27700"),
              Extent.from_bbox((100, 200, 300, 400), crs="EPSG:27700"),
              "suburban_area",
              'https://api.os.uk/search/names/v1/find?key=test',
              {"query": "Buckingham Palace", "offset": 0, "bounds": "100.0,200.0,300.0,400.0",
               "fq": ["LOCAL_TYPE:Suburban_Area", "BBOX:100.0,200.0,300.0,400.0"], "maxresults": 50}
              )
    ]
    return test_variables, test_data


def test_find_live():
    test_variables = "text, limit, expected_length, minimum_length"
    test_data = [
        param("AB22 9", 100, 100, None),
        param("AB22 9", 47, 47, None),
        param("AB22 9", 1000, None, 100),
    ]
    return test_variables, test_data


def test_find_fail():
    test_variables = "text, limit, bounds, bbox_filter, local_type, expected_result"
    test_data = [
        param("Buckingham Palace",
              -1,
              None,
              None,
              None,
              ValueError
              ),
        param("Buckingham Palace",
              50,
              Extent.from_bbox((100, 200, 300, 400), crs="EPSG:4326"),
              None,
              None,
              (TypeError, TypeCheckError)
              ),
        param("Buckingham Palace",
              50,
              None,
              Extent.from_bbox((100, 200, 300, 400), crs="EPSG:4326"),
              None,
              (TypeError,TypeCheckError)
              ),
        param("Buckingham Palace",
              50,
              None,
              None,
              "testtest",
              ValueError),
    ]
    return test_variables, test_data


def test_nearest_pass():
    test_variables = "point, radius, local_type, expected_url, expected_params"
    test_data = [
        param(
            (100, 200),
            100,
            None,
            'https://api.os.uk/search/names/v1/nearest?key=test',
            {"point": "100,200", "radius": 100}
        ),
        param(
            (1000, 2000),
            100,
            None,
            'https://api.os.uk/search/names/v1/nearest?key=test',
            {"point": "1000,2000", "radius": 100}
        ),
        param(
            (100, 200),
            500,
            None,
            'https://api.os.uk/search/names/v1/nearest?key=test',
            {"point": "100,200", "radius": 500}
        ),
        param(
            (100, 200),
            100,
            None,
            'https://api.os.uk/search/names/v1/nearest?key=test',
            {"point": "100,200", "radius": 100}
        ),
        param(
            (100, 200),
            100,
            "oil_refining",
            'https://api.os.uk/search/names/v1/nearest?key=test',
            {"point": "100,200", "radius": 100,
                "fq": ["LOCAL_TYPE:Oil_Refining"]}
        ),
        param(
            (100, 200),
            100,
            ("oil_refining", "oil_terminal"),
            'https://api.os.uk/search/names/v1/nearest?key=test',
            {"point": "100,200", "radius": 100, "fq": [
                "LOCAL_TYPE:Oil_Refining LOCAL_TYPE:Oil_Terminal"]}
        )
    ]
    return test_variables, test_data


def test_nearest_fail():
    test_variables = "point, radius, local_type, expected_result"
    test_data = [
        param(
            ("dog", "cat"),
            100,
            None,
            (TypeError, TypeCheckError)
        ),
        param(
            (100, 200),
            1001,
            None,
            ValueError
        ),
        param(
            (100, 200),
            -5,
            None,
            ValueError
        ),
        param(
            (100, 200),
            100,
            "testtest",
            ValueError
        ),
        param(
            (100, 200),
            100,
            ["suburban_area", "testest"],
            ValueError
        )
    ]
    return test_variables, test_data



from typing import Callable

from osdatahub.filters import (contains, crosses, disjoint, equals, intersects,
                               overlaps, touches, within)


class SpatialFilterTypes:
    contains = contains
    crosses = crosses
    disjoint = disjoint
    equals = equals
    intersects = intersects
    overlaps = overlaps
    touches = touches
    within = within

    @classmethod
    def get(cls, filter_type: str) -> Callable:
        """
        Get relevant spatial filter from string
        
        Args:
        filter_type (str): name of ogc gml filter

        Returns:
            Callable: A function that generates the desired geospatial XML filter
        """
        if hasattr(SpatialFilterTypes, filter_type):
            return getattr(SpatialFilterTypes, filter_type)
        else:
            raise ValueError(f"unrecognised spatial filter type: {filter_type}")



import os
import tempfile
import unittest.mock as mock
import dotenv

dotenv.load_dotenv()

import pytest
from osdatahub import OpenDataDownload, DataPackageDownload
from osdatahub.DownloadsAPI.downloads_api import _DownloadObj

from tests.data import downloads_data as data

API_KEY = os.environ.get("OSDATAHUB_TEST_KEY")


class TestOpenData:
    @pytest.fixture()
    def open_data_download(self):
        product = OpenDataDownload(product_id="test_id")
        yield product

    def test_product_init(self, open_data_download):
        assert open_data_download._ENDPOINT == "https://api.os.uk/downloads/v1/products"
        assert open_data_download.id == "test_id"

    @pytest.mark.parametrize(*data.product_list_pass("test_id"))
    @pytest.mark.usefixtures("open_data_download")
    @mock.patch('osdatahub.get')
    def test_product_list_pass(self, request_mocked, open_data_download, file_name, file_format, file_subformat, area,
                               return_downloadobj, expected_url, expected_params):
        open_data_download.product_list(file_name=file_name, file_format=file_format, file_subformat=file_subformat,
                                        area=area, return_downloadobj=return_downloadobj)
        request_mocked.assert_called_with(url=expected_url,
                                          params=expected_params,
                                          proxies={})

        assert type(return_downloadobj) == bool

    @pytest.mark.parametrize(*data.product_list_fail())
    def test_product_list_fail(self, open_data_download, file_name, file_format, file_subformat, area,
                               return_downloadobj, expected_result):
        with pytest.raises(expected_result):
            open_data_download.product_list(file_name=file_name, file_format=file_format,
                                            file_subformat=file_subformat,
                                            area=area, return_downloadobj=return_downloadobj)

    @pytest.mark.skipif(API_KEY is None, reason="Test API key not available")
    def test_product_list_live(self):
        greenspaces = OpenDataDownload("OpenGreenspace")
        greenspaces_info = greenspaces.product_list()
        assert type(greenspaces_info) is list
        assert type(greenspaces_info[0]) is dict

    # @pytest.mark.parametrize(*data.download_pass())
    # def test_download_pass(self, open_data_download, output_dir, download_multiple, overwrite, processes,
    #                        product_list_mock_return,
    #                        downloadobj_mock_return,
    #                        expected_download_return,
    #                        download_callded_value):
    #
    #     open_data_download.product_list = mock.Mock(return_value=product_list_mock_return)
    #     _DownloadObj.download = mock.Mock(return_value=downloadobj_mock_return)
    #     tqdm.tqdm = mock.Mock(return_value="tqdm")
    #     open_data_download = OpenDataDownload("OpenGreenspace")
    #     response = open_data_download.download(output_dir, download_multiple=download_multiple, overwrite=overwrite)
    #     assert response == expected_download_return
    #     _DownloadObj.download.assert_has_calls(download_called_value)

# TODO Fix Test for packages

# class TestDataPackage:
#     @pytest.fixture()
#     def data_package(self):
#         data_package = DataPackageDownload(key="test_key", product_id="test_id")
#         yield data_package

#     @pytest.mark.skipif(API_KEY is None, reason="Test API key not available")
#     def test_download_pass(self):
#         # Arrange
#         product_package = DataPackageDownload(API_KEY, "97")

#         # Act
#         with tempfile.TemporaryDirectory() as tmpdirname:
#             downloaded = product_package.download("17094", tmpdirname, "bld_fts_buildingpart_orderSummary.json")

#         # Assert
#         assert len(downloaded) == 1

#     def test_download_list_pass(self):
#         # TODO: implement download_list_pass
#         pass

#     def test_download_list_fail(self):
#         # TODO: implement download_list_fail
#         pass

#     def test_versions_pass(self):
#         # TODO: implement versions_pass
#         pass

#     def test_versions_fail(self):
#         # TODO: implement versions_fail
#         pass


# class TestDownloadObj:
#     @pytest.fixture()
#     def download_obj(self):
#         download_obj = _DownloadObj(url="test_url", file_name="test_file", size=256)
#         yield download_obj

#     def download_pass(self):
#         # TODO: implement _DownloadObj download pass
#         pass

#     def download_fail(self):
#         # TODO: implement _DownloadObje download fail
#         pass



from osdatahub.grow_list import GrowList


class TestGrowList:

    def test_extend(self):
        # Arrange
        gl = GrowList([1, 2, 3])

        # Act
        gl.extend([4, 5, 6])

        # Assert
        assert gl.values == [1, 2, 3, 4, 5, 6]

    def test_grown_true(self):
        # Arrange
        gl = GrowList([1, 2, 3])

        # Act
        gl.extend([4, 5, 6])

        # Assert
        assert gl.grown

    def test_grown_true_init(self):
        # Arrange
        gl = GrowList([1, 2, 3])

        # Assert
        assert gl.grown

    def test_grown_false(self):
        # Arrange
        gl = GrowList([1, 2, 3])
        gl.extend([])

        # Assert
        assert not gl.grown



from datetime import datetime

from osdatahub import Extent
from pytest import param
from shapely.geometry import Polygon


def test_ngd_query():
    test_variables = "extent, crs, start_datetime, end_datetime, cql_filter, filter_crs, max_results, offset, " \
                     "expected_url, expected_params"
    test_data = [
        param(
            None,
            None,
            None,
            None,
            None,
            None,
            100,
            0,
            r"https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items/",
            {
                "limit": 100,
                "offset": 0
            }
        ),
        param(
            None,
            None,
            None,
            None,
            None,
            None,
            50,
            0,
            r"https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items/",
            {
                "limit": 50,
                "offset": 0
            }
        ),
        param(
            None,
            None,
            None,
            None,
            None,
            None,
            100,
            20,
            r"https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items/",
            {
                "limit": 100,
                "offset": 20
            }
        ),
        param(Extent.from_bbox((-2.503510, 53.578646, -2.432785, 53.600655), crs="EPSG:4326"),
              None,
              None,
              None,
              None,
              None,
              100,
              0,
              r"https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items/",
              {
                  "limit": 100,
                  "offset": 0,
                  "filter": "INTERSECTS(geometry, POLYGON ((-2.432785 53.578646, -2.432785 53.600655, "
                            "-2.50351 53.600655, -2.50351 53.578646, -2.432785 53.578646)))",
                  "filter-crs": "http://www.opengis.net/def/crs/EPSG/0/4326"
              }
              ),
        param(Extent.from_bbox((-278689.4051, 7090757.3659, -270816.3912, 7094884.9654), crs="EPSG:3857"),
              None,
              None,
              None,
              None,
              None,
              100,
              0,
              r"https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items/",
              {
                  "limit": 100,
                  "offset": 0,
                  "filter": "INTERSECTS(geometry, POLYGON ((-270816.3912 7090757.3659, -270816.3912 7094884.9654, "
                            "-278689.4051 7094884.9654, -278689.4051 7090757.3659, -270816.3912 7090757.3659)))",
                  "filter-crs": "http://www.opengis.net/def/crs/EPSG/0/3857"
              }
              ),
        param(Extent(Polygon([[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]), crs="EPSG:3857"),
              None,
              None,
              None,
              None,
              None,
              100,
              0,
              r"https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items/",
              {
                  "limit": 100,
                  "offset": 0,
                  "filter": "INTERSECTS(geometry, POLYGON ((0 0, 1 0, 1 1, 0 1, 0 0)))",
                  "filter-crs": "http://www.opengis.net/def/crs/EPSG/0/3857"
              }
              ),
        param(
            None,
            3857,
            None,
            None,
            None,
            None,
            100,
            20,
            r"https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items/",
            {
                "limit": 100,
                "offset": 20,
                "crs": "http://www.opengis.net/def/crs/EPSG/0/3857"
            }
        ),
        param(
            None,
            3857,
            datetime(2014, 1, 1, 0, 0, 1),
            None,
            None,
            None,
            100,
            20,
            r"https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items/",
            {
                "limit": 100,
                "offset": 20,
                "crs": "http://www.opengis.net/def/crs/EPSG/0/3857",
                "datetime": "2014-01-01T00:00:01Z/.."
            }
        ),
        param(
            None,
            3857,
            None,
            datetime(2015, 1, 1, 0, 0, 1),
            None,
            None,
            100,
            20,
            r"https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items/",
            {
                "limit": 100,
                "offset": 20,
                "crs": "http://www.opengis.net/def/crs/EPSG/0/3857",
                "datetime": "../2015-01-01T00:00:01Z"
            }
        ),
        param(
            None,
            3857,
            datetime(2014, 1, 1, 0, 0, 1),
            datetime(2015, 1, 1, 0, 0, 1),
            None,
            None,
            100,
            20,
            r"https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items/",
            {
                "limit": 100,
                "offset": 20,
                "crs": "http://www.opengis.net/def/crs/EPSG/0/3857",
                "datetime": "2014-01-01T00:00:01Z/2015-01-01T00:00:01Z"
            }
        ),
        param(
            None,
            None,
            None,
            None,
            "HELLO WORLD",
            27700,
            100,
            0,
            r"https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items/",
            {
                "limit": 100,
                "offset": 0,
                "filter": "HELLO WORLD",
                "filter-crs": "http://www.opengis.net/def/crs/EPSG/0/27700"
            }
        ),
        param(Extent(Polygon([[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]), crs="EPSG:3857"),
              None,
              None,
              None,
              "HELLO WORLD",
              None,
              100,
              0,
              r"https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items/",
              {
                  "limit": 100,
                  "offset": 0,
                  "filter": "HELLO WORLD AND INTERSECTS(geometry, POLYGON ((0 0, 1 0, 1 1, 0 1, 0 0)))",
                  "filter-crs": "http://www.opengis.net/def/crs/EPSG/0/3857"
              }
              ),
        param(Extent(Polygon([[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]), crs="EPSG:3857"),
              None,
              None,
              None,
              "HELLO WORLD",
              "epsg:3857",
              100,
              0,
              r"https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items/",
              {
                  "limit": 100,
                  "offset": 0,
                  "filter": "HELLO WORLD AND INTERSECTS(geometry, POLYGON ((0 0, 1 0, 1 1, 0 1, 0 0)))",
                  "filter-crs": "http://www.opengis.net/def/crs/EPSG/0/3857"
              }
              ),
        param(Extent(Polygon([[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]), crs="EPSG:3857"),
              7405,
              datetime(2013, 1, 3, 12, 13, 14),
              datetime(2016, 3, 2, 1, 2, 3),
              "HELLO WORLD",
              "epsg:3857",
              100,
              0,
              r"https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items/",
              {
                  "limit": 100,
                  "offset": 0,
                  "crs": "http://www.opengis.net/def/crs/EPSG/0/7405",
                  "datetime": "2013-01-03T12:13:14Z/2016-03-02T01:02:03Z",
                  "filter": "HELLO WORLD AND INTERSECTS(geometry, POLYGON ((0 0, 1 0, 1 1, 0 1, 0 0)))",
                  "filter-crs": "http://www.opengis.net/def/crs/EPSG/0/3857"
              }
              ),
    ]

    return test_variables, test_data


def test_ngd_query_fail():
    test_variables = "extent, crs, start_datetime, end_datetime, cql_filter, filter_crs, max_results, offset, " \
                     "expected_error"
    test_data = [
        param(
            Extent(Polygon([[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]), crs="EPSG:3857"),
            None,
            None,
            None,
            "HELLO WORLD",
            "EPSG:4326",
            100,
            0,
            AssertionError
        ),
        param(
            None,
            None,
            datetime(2014, 1, 1, 0, 0, 1),
            datetime(2013, 1, 1, 0, 0, 1),
            None,
            None,
            100,
            0,
            ValueError
        ),
        param(
            None,
            None,
            None,
            None,
            None,
            None,
            0,
            0,
            AssertionError
        ),
        param(
            None,
            None,
            None,
            None,
            None,
            None,
            1,
            -1,
            AssertionError
        ),

    ]

    return test_variables, test_data


def test_ngd_instantiation():
    test_variables = "api_key, collection, expected_url, expected_headers"
    test_data = [
        param(
            "test_key",
            "test_collection",
            "https://api.os.uk/features/ngd/ofa/v1/collections/test_collection/items/",
            {"key": "test_key", 'Accept': 'application/geo+json'}
        ),
        param(
            "test_key2",
            "test_collection2",
            "https://api.os.uk/features/ngd/ofa/v1/collections/test_collection2/items/",
            {"key": "test_key2", 'Accept': 'application/geo+json'}
        )
    ]
    return test_variables, test_data


def test_ngd_query_feature():
    test_variables = "feature_id, crs, expected_url, expected_params"
    test_data = [
        param("feature_id",
              None,
              "https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items/feature_id",
              {}
              ),
        param("feature_id2",
              None,
              "https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items/feature_id2",
              {}
              ),
        param("feature_id2",
              3857,
              "https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items/feature_id2",
              {"crs": "http://www.opengis.net/def/crs/EPSG/0/3857"}
              ),
        param("feature_id2",
              "epsg:7405",
              "https://api.os.uk/features/ngd/ofa/v1/collections/bld-fts-buildingline/items/feature_id2",
              {"crs": "http://www.opengis.net/def/crs/EPSG/0/7405"}
              ),
    ]

    return test_variables, test_data



def test_ngd_query_live():
    test_variables = "collection, extent, crs, start_datetime, end_datetime, cql_filter, filter_crs, max_results, offset"
    test_data = [
        param(
            "bld-fts-buildingpart",
            None,
            None,
            None,
            None,
            None,
            None,
            5,
            0
        ),
        param(
            "lus-fts-site",
            None,
            None,
            None,
            None,
            None,
            None,
            50,
            0
        ),
        param(
            "str-fts-structureline",
            None,
            None,
            None,
            None,
            None,
            None,
            5,
            20
        ),
        param("trn-ntwk-road",
            Extent.from_bbox((-2.503510, 53.578646, -2.432785, 53.600655), crs="CRS84"),
              None,
              None,
              None,
              None,
              None,
              5,
              0
              ),
        param("trn-rami-highwaydedication",
            Extent.from_bbox((-278689.4051, 7090757.3659, -270816.3912, 7094884.9654), crs="EPSG:3857"),
              None,
              None,
              None,
              None,
              None,
              5,
              0
              ),
        param("wtr-ntwk-waterlink",
            Extent.from_bbox((280371.5367, 597854.9713, 320300.6732, 625998.9645), crs="EPSG:27700"),
              None,
              None,
              None,
              None,
              None,
              5,
              0
              ),
        param("trn-fts-rail",
            None,
            3857,
            None,
            None,
            None,
            None,
            7,
            20
        ),
        param("trn-ntwk-road",
            None,
            3857,
            None,
            None,
            None,
            None,
            134,
            20,
            id=">100 features"
        )
    ]

    return test_variables, test_data



import asyncio
import logging
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple, Union

from typeguard import typechecked

import osdatahub
from osdatahub import Extent
from osdatahub.AsyncAPI import AsyncHTTPClient
from osdatahub.NGD.crs import get_crs
from osdatahub.NGD.models import NGDFeatureCollection


# TODO: check that this is more efficient - avoids having to do the copy each time like the synchonous version
def _merge_all_geojsons(geojsons: List[Dict]) -> NGDFeatureCollection:
    """Merge multiple GeoJSON FeatureCollections into one efficiently."""
    if not geojsons:
        raise ValueError("No geojsons to merge")

    required_keys = {"features", "numberReturned", "links"}
    for gj in geojsons:
        if not gj.keys() >= required_keys:
            raise ValueError(f"All geojsons must contain keys {required_keys}")

    base = geojsons[0]
    features = list(base.get("features", []))
    links = list(base.get("links", []))
    count = base.get("numberReturned", 0)

    for gj in geojsons[1:]:
        features.extend(gj.get("features", []))
        links.extend(gj.get("links", []))
        count += gj.get("numberReturned", 0)

    return NGDFeatureCollection(
        type=base.get("type", "FeatureCollection"),
        features=features,
        numberReturned=count,
        links=links,
        timeStamp=base.get("timeStamp"),
        numberMatched=base.get("numberMatched"),
    )


class AsyncNGD:
    """
    Async client for querying OS NGD Features API.

    This class provides async methods for querying the OS National Geographic Database
    with support for parallel pagination, rate limiting, and automatic retries.

    Args:
        key: A valid OS Data Hub API key. Get a free key at https://osdatahub.os.uk/
        collection: ID for the desired NGD Feature Collection. See available collections at
            https://osdatahub.os.uk/docs/ofa/technicalSpecification
        max_concurrent: Maximum concurrent requests (default: 5)
        request_delay: Delay between requests in seconds (default: 0.3)
        max_retries: Maximum retry attempts on failure (default: 3)

    Example::

        from osdatahub import AsyncNGD, Extent
        import asyncio

        async def main():
            key = "your-api-key"
            extent = Extent.from_bbox((600000, 310200, 600900, 310900), "EPSG:27700")

            async with AsyncNGD(key, "trn-ntwk-street-1") as ngd:
                # Fetches 500 features in parallel (5 pages of 100)
                results = await ngd.query(max_results=500, extent=extent)
                print(f"Got {len(results['features'])} features")

        asyncio.run(main())
    """

    __ENDPOINT = r"https://api.os.uk/features/ngd/ofa/v1/collections"
    __HEADERS = {"Accept": "application/geo+json"}
    __PAGE_SIZE = 100  # NGD API max return per request

    _collections_cache: Optional[Dict] = None

    def __init__(
        self,
        key: str,
        collection: str,
        max_concurrent: int = 5,
        request_delay: float = 0.1,  # TODO: look at the OS throttling docs
        max_retries: int = 3,
    ) -> None:
        self.key: str = key
        self.collection: str = collection
        self._client: Optional[AsyncHTTPClient] = None

        # Store config
        self._max_concurrent = max_concurrent
        self._request_delay = request_delay
        self._max_retries = max_retries

    def _get_client(self) -> AsyncHTTPClient:
        """Initialisation of HTTP client."""
        if self._client is None:
            self._client = AsyncHTTPClient(
                max_concurrent=self._max_concurrent,
                request_delay=self._request_delay,
                max_retries=self._max_retries,
                proxies=osdatahub.get_proxies(),
            )
        return self._client

    def _endpoint(self, feature_id: Optional[str] = None) -> str:
        """Build endpoint URL."""
        return f"{self.__ENDPOINT}/{self.collection}/items/{feature_id if feature_id else ''}"

    def _build_headers(self) -> Dict[str, str]:
        """Build request headers."""
        return self.__HEADERS.copy()

    def _add_auth(self, params: Dict[str, Any]) -> Dict[str, Any]:
        """Add API key to query parameters."""
        params = params.copy()
        params["key"] = self.key
        return params

    @classmethod
    async def get_collections(cls, client: Optional[AsyncHTTPClient] = None) -> Dict:
        """
        Async retrieval of all OS NGD Feature Collections.

        Args:
            client: Optional AsyncHTTPClient to use. If not provided,
                creates a temporary client that will be closed after the request.

        Returns:
            Dict containing all Feature Collections with details for each.
        """
        if cls._collections_cache is not None:
            return cls._collections_cache

        should_close = client is None
        if client is None:
            client = AsyncHTTPClient(proxies=osdatahub.get_proxies())

        try:
            result = await client.get(cls.__ENDPOINT)
            cls._collections_cache = result
            return result
        finally:
            if should_close:
                await client.close()

    @classmethod
    def clear_collections_cache(cls) -> None:
        """Clear the cached collections data."""
        cls._collections_cache = None

    @typechecked
    async def query(
        self,
        extent: Optional[Extent] = None,
        crs: Optional[Union[str, int]] = None,
        start_datetime: Optional[datetime] = None,
        end_datetime: Optional[datetime] = None,
        cql_filter: Optional[str] = None,
        filter_crs: Optional[Union[str, int]] = None,
        max_results: Optional[int] = 100,
        offset: int = 0,
    ) -> NGDFeatureCollection:
        """
        Async query for features from a Collection with parallel pagination.

        This method will automatically paginate through results, fetching
        multiple pages in parallel for improved performance.

        Args:
            extent: An Extent object for spatial filtering. Only features within
                this extent will be returned. Supports EPSG:27700, EPSG:4326,
                EPSG:3857, and CRS84.
            crs: CRS for returned features. Can be "epsg:xxxx" format or EPSG code
                as int. Available: EPSG:27700, EPSG:4326, EPSG:7405, EPSG:3857, CRS84.
                Defaults to CRS84.
            start_datetime: Filter for features with temporal property after this time.
            end_datetime: Filter for features with temporal property before this time.
            cql_filter: A CQL format filter query. See OS docs for supported operators.
            filter_crs: CRS for the CQL query geometry. Must match extent CRS if both provided.
            max_results: Maximum number of features to return. Set to None to fetch all
                available features (default: 100).
            offset: Starting offset for pagination (default: 0).

        Returns:
            FeatureCollection with merged results.

        Raises:
            AssertionError: If max_results <= 0 or offset < 0.
            ValueError: If start_datetime > end_datetime or CRS validation fails.
        """

        if max_results is not None and max_results <= 0:
            raise ValueError(f"max_results must be > 0, got {max_results}")
        if offset < 0:
            raise ValueError(f"offset must be >= 0, got {offset}")

        params = self._build_params(
            extent=extent,
            crs=crs,
            start_datetime=start_datetime,
            end_datetime=end_datetime,
            cql_filter=cql_filter,
            filter_crs=filter_crs,
        )

        client = self._get_client()
        headers = self._build_headers()

        # If max_results is None, fetch all available features
        # TODO: Could just default to this
        if max_results is None:
            return await self._fetch_all(client, params, headers, offset)

        # Calculate pagination for fixed max_results
        page_offsets = self._calculate_page_offsets(max_results, offset)

        tasks = []
        for page_offset, page_limit in page_offsets:
            page_params = {**params, "limit": page_limit, "offset": page_offset}
            tasks.append(self._fetch_page(client, page_params, headers))

        results = await asyncio.gather(*tasks, return_exceptions=True)

        successful_results: List[Dict] = []
        for i, result in enumerate(results):
            if isinstance(result, BaseException):
                logging.error(
                    f"Page fetch failed for offset {page_offsets[i][0]}: {result}"
                )
                raise result
            page_result: Dict = result
            features_count = len(page_result.get("features", []))

            if features_count < page_offsets[i][1]:
                successful_results.append(page_result)
                break
            successful_results.append(page_result)

        # TODO: maybe raise exception instead?
        if not successful_results:
            return NGDFeatureCollection(
                type="FeatureCollection",
                features=[],
                numberReturned=0,
                links=[],
            )

        return _merge_all_geojsons(successful_results)

    def _calculate_page_offsets(
        self, max_results: int, start_offset: int
    ) -> List[Tuple[int, int]]:
        """
        Calculate (offset, limit) tuples for parallel fetching.

        Args:
            max_results: Total number of results desired
            start_offset: Starting offset

        Returns:
            List of (offset, limit) tuples for each page to fetch
        """
        pages = []
        remaining = max_results
        current_offset = start_offset

        while remaining > 0:
            page_limit = min(remaining, self.__PAGE_SIZE)
            pages.append((current_offset, page_limit))
            current_offset += page_limit
            remaining -= page_limit

        return pages

    async def _fetch_page(
        self, client: AsyncHTTPClient, params: Dict, headers: Dict
    ) -> Dict:
        """Fetch a single page of results."""
        return await client.get(
            self._endpoint(), params=self._add_auth(params), headers=headers
        )

    async def _fetch_all(
        self,
        client: AsyncHTTPClient,
        params: Dict,
        headers: Dict,
        start_offset: int = 0,
    ) -> NGDFeatureCollection:
        """
        Fetch all available features by paginating until no more results.
        """
        all_results: List[Dict] = []
        current_offset = start_offset
        batch_size = self._max_concurrent

        while True:
            tasks = []
            offsets = []
            for i in range(batch_size):
                offset = current_offset + (i * self.__PAGE_SIZE)
                offsets.append(offset)
                page_params = {**params, "limit": self.__PAGE_SIZE, "offset": offset}
                tasks.append(self._fetch_page(client, page_params, headers))

            results = await asyncio.gather(*tasks, return_exceptions=True)

            done = False
            for i, result in enumerate(results):
                if isinstance(result, BaseException):
                    logging.error(
                        f"Page fetch failed for offset {offsets[i]}: {result}"
                    )
                    raise result

                page_result: Dict = result
                features_count = len(page_result.get("features", []))

                if features_count > 0:
                    all_results.append(page_result)

                if features_count < self.__PAGE_SIZE:
                    done = True
                    break

            if done:
                break

            current_offset += batch_size * self.__PAGE_SIZE

        # TODO: Raise exception here?
        if not all_results:
            return NGDFeatureCollection(
                type="FeatureCollection",
                features=[],
                numberReturned=0,
                links=[],
            )

        return _merge_all_geojsons(all_results)

    def _build_params(
        self,
        extent: Optional[Extent],
        crs: Optional[Union[str, int]],
        start_datetime: Optional[datetime],
        end_datetime: Optional[datetime],
        cql_filter: Optional[str],
        filter_crs: Optional[Union[str, int]],
    ) -> Dict[str, Any]:
        """Build query parameters (mirrors sync implementation logic)."""
        params: Dict[str, Any] = {}

        if crs:
            params["crs"] = get_crs(crs=crs)

        if start_datetime or end_datetime:
            if start_datetime and end_datetime and start_datetime > end_datetime:
                raise ValueError("Start time must be before end time")

            start_str = start_datetime.isoformat() + "Z" if start_datetime else ".."
            end_str = end_datetime.isoformat() + "Z" if end_datetime else ".."
            params["datetime"] = f"{start_str}/{end_str}"

        if extent:
            bbox_filter = f"INTERSECTS(geometry, {extent.polygon.wkt})"

            if cql_filter:
                if filter_crs:
                    assert get_crs(
                        extent.crs,
                        valid_crs=("epsg:4326", "epsg:27700", "epsg:3857", "crs84"),
                    ) == get_crs(filter_crs), "filter_crs must match extent crs"
                else:
                    filter_crs = extent.crs
                cql_filter += f" AND {bbox_filter}"
            else:
                cql_filter = bbox_filter
                filter_crs = extent.crs

        if cql_filter:
            params["filter"] = cql_filter
            if filter_crs:
                params["filter-crs"] = get_crs(
                    crs=filter_crs,
                    valid_crs=("epsg:4326", "epsg:27700", "epsg:3857", "crs84"),
                )

        return params

    async def query_feature(
        self, feature_id: str, crs: Optional[Union[str, int]] = None
    ) -> Dict:
        """
        Async retrieval of a single feature by ID.

        Args:
            feature_id: The unique identifier of the feature to retrieve.
            crs: CRS for the returned feature. Can be "epsg:xxxx" format or
                EPSG code as int. Defaults to CRS84.

        Returns:
            GeoJSON Feature object.
        """
        params = {"crs": get_crs(crs)} if crs else {}
        headers = self._build_headers()

        client = self._get_client()
        return await client.get(
            self._endpoint(feature_id), params=self._add_auth(params), headers=headers
        )

    async def close(self) -> None:
        """Close the HTTP client and release resources."""
        if self._client is not None:
            await self._client.close()
            self._client = None

    async def __aenter__(self) -> "AsyncNGD":
        """Async context manager entry."""
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        """Async context manager exit - cleanup."""
        await self.close()



from pytest import param


def test_get_endpoint():
    test_variables = "key, identifier, feature_type, identifier_type, expected_result"
    test_data = [
        param(
            "KEY",
            "IDENTIFIER",
            None,
            "TOID",
            "https://api.os.uk/search/links/v1/identifierTypes/TOID/IDENTIFIER?key=KEY",
            id="identifier_type only",
        ),
        param(
            "KEY",
            "IDENTIFIER",
            "Road",
            None,
            "https://api.os.uk/search/links/v1/featureTypes/Road/IDENTIFIER?key=KEY",
            id="feature_type only",
        ),
        param(
            "KEY",
            "IDENTIFIER",
            None,
            None,
            "https://api.os.uk/search/links/v1/identifiers/IDENTIFIER?key=KEY",
            id="no types",
        ),
    ]
    return test_variables, test_data



from collections.abc import Iterable
from typing import Union

from typeguard import typechecked

import osdatahub
from osdatahub.errors import raise_http_error
from osdatahub.extent import Extent
from osdatahub.grow_list import GrowList
from osdatahub.NamesAPI.local_types import get_local_type, validate_local_type
from osdatahub.utils import addresses_to_geojson


class NamesAPI:
    """Main class for querying the OS Names API (https://osdatahub.os.uk/docs/names/overview)

    Args:
        key (str): A valid OS API Key. Get a free key here - https://osdatahub.os.uk/

    Example::

        from osdatahub import NamesAPI
        from os import environ

        key = environ.get("OS_API_KEY")
        names = NamesAPI(key)
        results = names.find("Buckingham Palace", limit=5)
    """
    __ENDPOINT = r"https://api.os.uk/search/names/v1/"
    HEADERS = {"method": "POST",
               "headers": "{'Content-Type': 'application/json'}"}

    def __init__(self, key: str):
        self.key = key

    def __endpoint(self, api_name: str) -> str:
        return self.__ENDPOINT + api_name + f"?key={self.key}"

    @typechecked
    def find(self,
             text: str,
             limit: int = 100,
             bounds: Union[Extent, None] = None,
             bbox_filter: Union[Extent, None] = None,
             local_type: Union[Iterable, str, None] = None) -> dict:
        """A free text query of the OS Names API

        Args:
            text (str): The free text search parameter
            limit (int, optional): The maximum number of features to return.
                Defaults to 100.
            bounds (Extent, optional): Biases the results to a certain area. Must be British National Grid
                (EPSG:27700) CRS
            bbox_filter (Extent, optional): Filters the results to a certain area. Must be British National Grid
                (EPSG:27700) CRS
            local_type (Union[Iterable, str], optional): Filters the results to certain local types. Available local
                types can be found at the bottom of https://osdatahub.os.uk/docs/names/technicalSpecification

        Returns:
            FeatureCollection: The results of the query in GeoJSON format
        """
        data = GrowList()
        params = {"query": text}

        if limit <= 0:
            raise ValueError(f"Parameter \"limit\" must be a positive integer. Instead got {limit}")

        if bounds:
            if not bounds.crs == "EPSG:27700":
                raise TypeError("Bounds must be in British National Grid CRS (EPSG:27700)")
            params.update({"bounds": bounds.bbox.to_string(precision=2)})
        if bbox_filter or local_type:
            if bbox_filter and (not bbox_filter.crs == "EPSG:27700"):
                raise TypeError("Bounding Box filter must be in British National Grid CRS (EPSG:27700)")
            params.update({"fq": self.__format_fq(bbox_filter, local_type)})

        try:
            n_required = min(limit, 100)
            while n_required > 0 and data.grown:
                params.update({"offset": len(data), "maxresults": n_required})
                response = osdatahub.get(self.__endpoint("find"), params=params, proxies=osdatahub.get_proxies())
                data.extend(self.__format_response(response))
                n_required = min(100, limit - len(data))
        except KeyError:
            raise_http_error(response)
        return addresses_to_geojson(data.values, "EPSG:27700")

    @typechecked
    def nearest(self,
                point: tuple,
                radius: float = 100,
                local_type: Union[Iterable, str, None] = None) -> dict:
        """Takes a pair of coordinates (X, Y) as an input
        to determine the closest name.

        Args:
            point (tuple): A set of coordinates in British National Grid (EPSG:27700) format
            radius (float): The search radius in metres (min. 0.01, max. 1000).
                Defaults to 100.
            local_type (Union[Iterable, str], optional):  Filters the results to certain local types. Available local
                types can be found at the bottom of https://osdatahub.os.uk/docs/names/technicalSpecification

        Returns:
            FeatureCollection: The results of the query in GeoJSON format
        """
        data = GrowList()
        if not all([str(p).isnumeric() for p in point]):
            raise TypeError("All values in argument \"point\" must be numeric")
        if not 0.01 <= radius <= 1000:
            raise ValueError(f"Argument \"radius\" must be between 0.01 and 1000, but had value {radius}")

        params = {"point": ",".join([str(c) for c in point]), "radius": radius}
        if local_type:
            params.update({"fq": self.__format_fq(local_type=local_type)})
        try:
            response = osdatahub.get(self.__endpoint("nearest"), params=params, proxies=osdatahub.get_proxies())
            data.extend(self.__format_response(response))
        except KeyError:
            if response.status_code != 200:
                raise_http_error(response)
        return addresses_to_geojson(data.values, crs="EPSG:27700")

    @staticmethod
    @typechecked
    def __format_fq(bbox_filter: Union[Extent, None] = None,
                    local_type: Union[str, Iterable, None] = None) -> list:
        """
        Formats optional fq arguments for Names API query

        Args:
            bbox_filter (Extent, optional): Filters the results to a certain area. Must be British National Grid
                (EPSG:27700) CRS
            local_type (Union[str, Iterable], optional): Filters the results to certain local types. Available local
                types can be found at the bottom of https://osdatahub.os.uk/docs/names/technicalSpecification

        Returns:
            list of fq filtering arguments
        """
        fq_args = []
        if local_type:
            # check that all given local types are valid
            invalid_local_types = validate_local_type(local_type)
            if invalid_local_types:
                raise ValueError(f"The local type(s) {invalid_local_types} are not valid local types")
            # builds local_type query whether given one argument or multiple
            if isinstance(local_type, str):
                local_types = "LOCAL_TYPE:" + get_local_type(local_type)
            elif isinstance(local_type, Iterable):
                local_types = " ".join([f"LOCAL_TYPE:{get_local_type(arg)}" for arg in local_type])
            else:
                raise TypeError(
                    f"'local_type' argument must be Iterable or str, but was type {type(local_type)}")
            fq_args.append(local_types)

        # adds bbox filter to argument
        if bbox_filter:
            if not bbox_filter.crs == "EPSG:27700":
                raise ValueError("'bbox_filter' argument must have CRS of British National Grid (EPSG:27700). Its CRS "
                                 f"is {bbox_filter.crs}")
            fq_args.append("BBOX:" + str(bbox_filter.bbox.to_string(precision=2)))

        return fq_args

    @staticmethod
    def __format_response(response) -> list:
        try:
            return [result["GAZETTEER_ENTRY"] for result
                    in response.json()["results"]]
        except KeyError as e:
            if response.status_code == 200:
                    return []
            raise e


from dataclasses import dataclass
from typing import Collection, Union, Iterable

import osdatahub
from osdatahub.bbox import BBox
from osdatahub.ons_api import get_ons_geom
from shapely.geometry import Polygon, box
from shapely.geometry.point import Point

_VALID_CRS = ("EPSG:27700", "EPSG:3857", "EPSG:4326", "EPSG:7405", "CRS84")


@dataclass(frozen=True)
class Extent:
    """
    Extent is a dataclass that has two main uses:

        - It is a standardised way to store polygon extents that will interface
          with all of the osdatahub APIs.
        - It has multiple constructors that allow the user to create custom
          Extents without needing to write out polygon coordinates.

    """

    polygon: Polygon
    crs: str

    def __post_init__(self):
        if not isinstance(self.polygon, Polygon):
            raise TypeError("Extent expects geometry as a shapley Polygon")
        if self.crs.upper() not in _VALID_CRS:
            raise ValueError(
                f"Extent CRS must be one of {_VALID_CRS}, got '{self.crs}'"
            )

    @property
    def bbox(self) -> BBox:
        return BBox(*self.polygon.bounds)

    @property
    def xml_coords(self) -> str:
        coords = self.polygon.exterior.coords
        if self.crs.upper() == "EPSG:4326":
            return " ".join([f"{c2},{c1}" for c1, c2 in coords])
        return " ".join([f"{c1},{c2}" for c1, c2 in coords])

    @classmethod
    def from_bbox(cls, bbox: Union[Collection[float], BBox], crs: str) -> "Extent":
        """Creates a rectangular extent, given a bounding box.

        Args:
            bbox (Union[Collection[float], BBox]): A bounding box, passed in as either a
                BBox object or a collection of the form (west, south, east, north).
            crs (str): The CRS corresponding to the bouding box, must be either
                ''EPSG:4326', EPSG:27700' or 'EPSG:3857'.

        Raises:
            TypeError: If bbox is not a tuple or BBox object.

        Returns:
            Extent: A rectangular extent that matches
            the specified bounding box.
        """
        try:
            return Extent(box(*bbox), crs)
        except TypeError:
            raise TypeError(
                "bbox must be a BBox object or a collection "
                "of the form (west, south, east, north)"
            ) from None

    @classmethod
    def from_radius(cls, centre: Iterable, radius: float, crs: str) -> "Extent":
        """Creates a circular extent, given a centre point and a radius.

        Args:
            centre (Iterable): Either a coordinate tuple in the form (x, y)
                or a shapely Point.
            radius (float): The radius of the circle in metres
            crs (str): The CRS corresponding to the point coordinate,
                must be either 'EPSG:27700' or 'EPSG:3857'.

        Raises:
            TypeError: If centre is not a tuple or shapely Point.

        Returns:
            Extent: A rectangular extent that matches
            the specified bounding box.
        """
        if crs.upper() not in ("EPSG:27700", "EPSG:3857"):
            raise ValueError(
                f"crs must be one of ('EPSG:27700', 'EPSG:3857'), got {crs}"
            )
        if isinstance(centre, Iterable) and len(centre) == 2:
            return Extent(Point(*centre).buffer(radius), crs)
        elif isinstance(centre, Point):
            return Extent(centre.buffer(radius), crs)
        else:
            raise TypeError(
                "centre must be a shapely Point object "
                "or a coordinate tuple of the form (x, y)"
            )

    def set_crs(self, crs: str) -> "Extent":
        """
        Changes the coordinate reference system of the Extent object to the value specified

        Args:
            crs (str): The CRS to change to, must be either ‘’EPSG:4326’, EPSG:27700’ or ‘EPSG:3857’.

        Returns:
            Extent: A rectangular extent with an updated CRS
        """
        return Extent(self.polygon, crs)

    def is_within(self, bbox: Union[Collection[float], BBox]) -> bool:
        """
        Checks whether a bounding box is within the Extent object

        Args:
            bbox (Union[Collection[float], BBox]): A bounding box, passed in as either a
                BBox object or a collection of the form (west, south, east, north).

        Returns:
            bool: True if the bounding box is within the Extent object, False if not
        """
        try:
            bbox = BBox(*bbox)
        except TypeError:
            raise TypeError(
                "bbox must be a BBox object or a collection "
                "of the form (west, south, east, north)"
            ) from None
        return bbox[:2] <= self.bbox[:2] and bbox[-2:] >= self.bbox[-2:]

    def to_json(self):
        """
        Converts the extent object into a json

        Returns:
            dict: A json containing the coordinates of the Extent
        """
        coords = list(self.polygon.exterior.coords)
        if self.crs.upper() == "EPSG:4326":
            coords = [(c2, c1) for c1, c2 in coords]
        return {"type": "Polygon", "coordinates": [coords]}

    @classmethod
    def from_ons_code(cls, ons_code: str) -> "Extent":
        """Creates an extent of an existing ONS boundary.
        Note that the output will be in the "EPSG:4326" coordinate system.
        A full list of available ONS geographies can be found here:
        http://statistics.data.gov.uk/atlas/resource?uri=http://statistics.data.gov.uk/id/statistical-geography/K02000001

        Args:
            ons_code (str): The code for the desired geography
        Raises:
            ValueError: If the ONS geography is not a Polygon.

        Returns:
            Extent: An extent of the ONS geography.
        """
        geom = get_ons_geom(ons_code)
        geom_type = geom["type"]
        if geom_type != "Polygon":
            raise ValueError(
                f"osdatahub doesn't currently support geometry types other than Polygon. \n The ONS geography {ons_code} is a {geom_type}"
            )
        polygon = Polygon(geom["coordinates"][0])
        return Extent(polygon, "EPSG:4326")

    def __repr__(self):
        coords = list(self.polygon.exterior.coords)
        return (
            f"{self.__class__.__name__}(polygon=Polygon({coords})," f"crs='{self.crs}')"
        )

    def __eq__(self, o: object) -> bool:
        return self.polygon.equals(o.polygon) and (self.crs == o.crs)



import json

from requests.exceptions import HTTPError


RESPONSE_CODES = {
    200: ("OK", "Request has been successful."),
    304: (
        "Not Modified",
        "In response to a conditional GET request this response indicates that the underlying data has not changed since the previous request, and cached results may be re-used.",
    ),
    400: ("Bad request", "E.g. missing query parameter, malformed syntax."),
    401: (
        "Unauthorized",
        "API key is either incorrect, or does not permit access to the chosen product",
    ),
    403: (
        "Forbidden",
        "The client has authenticated its access but does not have sufficient rights to complete the request.",
    ),
    404: ("Not found", "The server has not found anything matching the Request-URI."),
    405: (
        "Method not allowed",
        "Request used an unsupported HTTP method, e.g. DELETE or PUT.",
    ),
    429: (
        "Too many requests",
        "Exceeded the number of requests per minute (rate-limit).",
    ),
    500: ("Internal server error", "Generic internal server error."),
    503: ("Service unavailable", "Temporary outage due to overloading or maintenance."),
}


def raise_http_error(response):
    code = response.status_code
    if code in RESPONSE_CODES:
        info = RESPONSE_CODES[code]
        url = response.url
        error_str = f"Unsuccessful query: {url}\n\nError {code} - {info[0]}: {info[1]}"
    else:
        try:
            is_too_large = response.json()["fault"]["detail"]["errorcode"] == "protocol.http.TooBigLine"
        except (KeyError, json.decoder.JSONDecodeError):
            is_too_large = False

        if is_too_large:
            error_str = "Query is too large. Please try simplifying it (e.g. reducing the vertices of a polygon)."
        else:
            error_str = response.text
    raise HTTPError(error_str)



from dataclasses import astuple, dataclass
from typing import Union


@dataclass
class BBox:
    """
    BBox is a dataclass that specifies a rectangular polygon made up of north, south, east and west. It is indiscriminate
    of CRS and so is used as part of the more comprehensive Extent class
    """

    west: float
    south: float
    east: float
    north: float

    def __iter__(self):
        return (c for c in astuple(self))

    def __getitem__(self, index):
        return list(self)[index]

    def to_string(self, precision: Union[int, None] = None) -> str:
        """
        Converts bounding box into string

        Args:
            precision (int): Decimal point rounding precision. Defaults to None (no rounding)

        Returns:
            str: bounding box in string form
        """
        return ",".join(str(round(c, precision) if precision is not None else c) for c in self)



from typing import Union

import requests
from typeguard import typechecked

import osdatahub
from osdatahub.errors import raise_http_error
from osdatahub.LinkedIdentifiersAPI.linked_identifier_options import (
    correlation_methods, feature_types, identifier_types)


class LinkedIdentifiersAPI:
    """Main class for querying the OS Linked Identifiers API (https://osdatahub.os.uk/docs/linkedIdentifiers/overview)

    Args:
        key (str): A valid OS API Key. Get a free key here - https://osdatahub.os.uk/

    Example::

        from osdatahub import LinkedIdentifiersAPI
        from os import environ

        key = environ.get("OS_API_KEY")
        linked_ids = LinkedIdentifiersAPI(key)
        results = linked_ids.query(200001025758)
    """

    __ENDPOINT = r"https://api.os.uk/search/links/v1/"

    def __init__(self, key: str):
        self.key = key

    @staticmethod
    def __request(endpoint: str) -> dict:
        response = osdatahub.get(endpoint, proxies=osdatahub.get_proxies())
        if response.status_code != 200:
            raise_http_error(response)
        return response.json()

    def __get_endpoint(
            self, identifier: Union[int, str], feature_type: str, identifier_type: str
    ) -> str:
        if feature_type is not None and identifier_type is not None:
            raise ValueError(
                "It is possible to query by the feature_type "
                + "OR the identifier type, but not both"
            )
        elif feature_type is not None:
            feature_types.validate(feature_type)
            subdirectory = f"featureTypes/{feature_type}"
        elif identifier_type is not None:
            identifier_types.validate(identifier_type)
            subdirectory = f"identifierTypes/{identifier_type}"
        else:
            subdirectory = "identifiers"
        return self.__ENDPOINT + subdirectory + f"/{identifier}?key={self.key}"

    @typechecked
    def query(
            self,
            identifier: Union[int, str],
            feature_type: Union[str, None] = None,
            identifier_type: Union[str, None] = None,
    ) -> dict:
        """Run a query of the OS Linked Identifiers API - looks up an
        identifier and finds its associated identifiers.

        Queries can be also be made more specific if the feature type
        or identifier type are known, but note: you cannot specify both!

        Args:
            identifier (Union[int, str]): The identifier to look up.
            feature_type (str): Look up linked identifiers when the input feature type is known.
            identifier_type (str): Look up linked identifiers when the input identifier type is known.

        Returns:
            dict: The results of the query in JSON format
        """
        endpoint = self.__get_endpoint(identifier, feature_type, identifier_type)
        return self.__request(endpoint)

    @typechecked
    def product_version(self, correlation_method: str) -> dict:
        """Discover the current product version information. For a list of
        valid correlation methods see (https://osdatahub.os.uk/docs/linkedIdentifiers/technicalSpecification)

        Args:
            correlation_method (str): Correlation method - corresponding to a
            particular feature relationship

        Returns:
            dict: The results of the query in JSON format
        """
        correlation_methods.validate(correlation_method)
        endpoint = (
                self.__ENDPOINT + f"productVersionInfo/{correlation_method}?key={self.key}"
        )
        return self.__request(endpoint)



from collections import namedtuple


Product = namedtuple("Product", "name geometry")


PREMIUM = {
    "topographic_area": Product("Topography_TopographicArea", "Polygon"),
    "topographic_point": Product("Topography_TopographicPoint", "Point"),
    "topographic_line": Product("Topography_TopographicLine", "LineString"),
    "water_network_link": Product("WaterNetwork_WatercourseLink", "LineString"),
    "water_network_node": Product("WaterNetwork_HydroNode", "Point"),
    "path_network_link": Product("DetailedPathNetwork_RouteLink", "LineString"),
    "path_network_node": Product("DetailedPathNetwork_RouteNode", "Point"),
    "highways_connecting_link": Product("Highways_ConnectingLink", "LineString"),
    "highways_connecting_node": Product("Highways_ConnectingNode", "Point"),
    "highways_ferry_link": Product("Highways_FerryLink", "LineString"),
    "highways_ferry_node": Product("Highways_FerryNode", "Point"),
    "highways_path_link": Product("Highways_PathLink", "LineString"),
    "highways_path_node": Product("Highways_PathNode", "Point"),
    "highways_road_link": Product("Highways_RoadLink", "LineString"),
    "highways_road_node": Product("Highways_RoadNode", "Point"),
    "highways_street": Product("Highways_Street", "LineString"),
    "greenspace_area": Product("Greenspace_GreenspaceArea", "Polygon"),
    "sites_access_point": Product("Sites_AccessPoint", "Point"),
    "sites_functional_site": Product("Sites_FunctionalSite", "Polygon"),
    "sites_routing_point": Product("Sites_RoutingPoint", "Point"),
    "topographic_boundary": Product("Topography_BoundaryLine", "LineString"),
    "cartographic_symbol": Product("Topography_CartographicSymbol", "Point"),
    "cartographic_text": Product("Topography_CartographicText", "Point"),

    "Topography_TopographicArea": Product("Topography_TopographicArea", "Polygon"),
    "Topography_TopographicPoint": Product("Topography_TopographicPoint", "Point"),
    "Topography_TopographicLine": Product("Topography_TopographicLine", "LineString"),
    "WaterNetwork_WatercourseLink": Product("WaterNetwork_WatercourseLink", "LineString"),
    "WaterNetwork_HydroNode": Product("WaterNetwork_HydroNode", "Point"),
    "DetailedPathNetwork_RouteLink": Product("DetailedPathNetwork_RouteLink", "LineString"),
    "DetailedPathNetwork_RouteNode": Product("DetailedPathNetwork_RouteNode", "Point"),
    "Highways_ConnectingLink": Product("Highways_ConnectingLink", "LineString"),
    "Highways_ConnectingNode": Product("Highways_ConnectingNode", "Point"),
    "Highways_FerryLink": Product("Highways_FerryLink", "LineString"),
    "Highways_FerryNode": Product("Highways_FerryNode", "Point"),
    "Highways_PathLink": Product("Highways_PathLink", "LineString"),
    "Highways_PathNode": Product("Highways_PathNode", "Point"),
    "Highways_RoadLink": Product("Highways_RoadLink", "LineString"),
    "Highways_RoadNode": Product("Highways_RoadNode", "Point"),
    "Highways_Street": Product("Highways_Street", "LineString"),
    "Greenspace_GreenspaceArea": Product("Greenspace_GreenspaceArea", "Polygon"),
    "Sites_AccessPoint": Product("Sites_AccessPoint", "Point"),
    "Sites_FunctionalSite": Product("Sites_FunctionalSite", "Polygon"),
    "Sites_RoutingPoint": Product("Sites_RoutingPoint", "Point"),
    "Topography_BoundaryLine": Product("Topography_BoundaryLine", "LineString"),
    "Topography_CartographicSymbol": Product("Topography_CartographicSymbol", "Point"),
    "Topography_CartographicText": Product("Topography_CartographicText", "Point")
}


OPEN = {
    "zoomstack_district_buildings": Product("Zoomstack_DistrictBuildings", "Polygon"),
    "zoomstack_foreshore": Product("Zoomstack_Foreshore", "Polygon"),
    "zoomstack_greenspace": Product("Zoomstack_Greenspace", "Polygon"),
    "zoomstack_local_buildings": Product("Zoomstack_LocalBuildings", "Polygon"),
    "zoomstack_national_parks": Product("Zoomstack_NationalParks", "Polygon"),
    "zoomstack_sites": Product("Zoomstack_Sites", "Polygon"),
    "zoomstack_surface_water": Product("Zoomstack_Surfacewater", "Polygon"),
    "zoomstack_urban_areas": Product("Zoomstack_UrbanAreas", "Polygon"),
    "zoomstack_woodland": Product("Zoomstack_Woodland", "Polygon"),
    "zoomstack_boundaries": Product("Zoomstack_Boundaries", "LineString"),
    "zoomstack_contours": Product("Zoomstack_Contours", "LineString"),
    "zoomstack_ETL": Product("Zoomstack_ETL", "LineString"),
    "zoomstack_rail": Product("Zoomstack_Rail", "LineString"),
    "zoomstack_roads_local": Product("Zoomstack_RoadsLocal", "LineString"),
    "zoomstack_roads_national": Product("Zoomstack_RoadsNational", "LineString"),
    "zoomstack_roads_regional": Product("Zoomstack_RoadsRegional", "LineString"),
    "zoomstack_waterlines": Product("Zoomstack_Waterlines", "LineString"),
    "open_USRN": Product("OpenUSRN_USRN", "LineString"),
    "zoomstack_airports": Product("Zoomstack_Airports", "Point"),
    "zoomstack_names": Product("Zoomstack_Names", "Point"),
    "zoomstack_railway_stations": Product("Zoomstack_RailwayStations", "Point"),
    "openUPRN_address": Product("OpenUPRN_Address", "Point"),
    "openTOID_highways_network": Product("OpenTOID_HighwaysNetwork", "Point"),
    "openTOID_sites": Product("OpenTOID_SitesLayer", "Point"),
    "openTOID_topography": Product("OpenTOID_TopographyLayer", "Point"),
    "Zoomstack_DistrictBuildings": Product("Zoomstack_DistrictBuildings", "Polygon"),
    "Zoomstack_Foreshore": Product("Zoomstack_Foreshore", "Polygon"),
    "Zoomstack_Greenspace": Product("Zoomstack_Greenspace", "Polygon"),
    "Zoomstack_LocalBuildings": Product("Zoomstack_LocalBuildings", "Polygon"),
    "Zoomstack_NationalParks": Product("Zoomstack_NationalParks", "Polygon"),
    "Zoomstack_Sites": Product("Zoomstack_Sites", "Polygon"),
    "Zoomstack_Surfacewater": Product("Zoomstack_Surfacewater", "Polygon"),
    "Zoomstack_UrbanAreas": Product("Zoomstack_UrbanAreas", "Polygon"),
    "Zoomstack_Woodland": Product("Zoomstack_Woodland", "Polygon"),
    "Zoomstack_Boundaries": Product("Zoomstack_Boundaries", "LineString"),
    "Zoomstack_Contours": Product("Zoomstack_Contours", "LineString"),
    "Zoomstack_ETL": Product("Zoomstack_ETL", "LineString"),
    "Zoomstack_Rail": Product("Zoomstack_Rail", "LineString"),
    "Zoomstack_RoadsLocal": Product("Zoomstack_RoadsLocal", "LineString"),
    "Zoomstack_RoadsNational": Product("Zoomstack_RoadsNational", "LineString"),
    "Zoomstack_RoadsRegional": Product("Zoomstack_RoadsRegional", "LineString"),
    "Zoomstack_Waterlines": Product("Zoomstack_Waterlines", "LineString"),
    "OpenUSRN_USRN": Product("OpenUSRN_USRN", "LineString"),
    "Zoomstack_Airports": Product("Zoomstack_Airports", "Point"),
    "Zoomstack_Names": Product("Zoomstack_Names", "Point"),
    "Zoomstack_RailwayStations": Product("Zoomstack_RailwayStations", "Point"),
    "openUPRN_address": Product("OpenUPRN_Address", "Point"),
    "OpenTOID_HighwaysNetwork": Product("OpenTOID_HighwaysNetwork", "Point"),
    "OpenTOID_SitesLayer": Product("OpenTOID_SitesLayer", "Point"),
    "OpenTOID_TopographyLayer": Product("OpenTOID_TopographyLayer", "Point")
}


def convert_product_to_new_geometry(product: Product) -> Product:
    """
    Takes a Product object and returns a copy with an updated geometry for the new API endpoint.
    The new API returns a MultiPolygon and a MultiLineString where the old API returned a Polygon and Linestring
    respectively.
    Args:
        product Product): Product object to convert

    Returns (Product): a copy of the Product object with updated geometry.
    """
    if product.geometry == "LineString":
        new_geom = "MultiLineString"
    elif product.geometry == "Polygon":
        new_geom = "MultiPolygon"
    elif product.geometry in ("Point", "MultiPolygon", "MultiLineString"):
        new_geom = product.geometry
    else:
        raise ValueError(f"argument product has an invalid geometry. Must be one of Point, Polygon, LineString, "
                         f"MultiPolygon, MultiLineString, but was {product.geometry}")

    return Product(product.name, new_geom)


OPEN_NEW = {key: convert_product_to_new_geometry(value) for key, value in OPEN.items()}

PREMIUM_NEW = {key: convert_product_to_new_geometry(value) for key, value in PREMIUM.items()}


def suggest_product(text: str) -> list:
    matches = []
    for product_name in OPEN:
        if text in product_name:
            matches.append(f"{product_name} [OPEN]")
    for product_name in PREMIUM:
        if text in product_name:
            matches.append(f"{product_name} [PREMIUM]")
    return matches


def validate_product_name(product_name: str) -> str:
    if product_name in OPEN or product_name in PREMIUM:
        return product_name
    suggested_products = suggest_product(product_name)
    suggestion_str = ", ".join(suggested_products) \
        if len(suggested_products) > 0 else "Can't find a match..."
    raise ValueError(f"Unrecognised product '{product_name}'.\n\n" \
                     f"\tBest Matches: {suggestion_str}\n\n" \
                     f"\tOpen Products: {', '.join(list(OPEN))}\n\n" \
                     f"\tPremium Products: {', '.join(list(PREMIUM))}\n\n")


def get_product(product_name: str, new_api: bool = False) -> Product:
    premium_lookup = PREMIUM_NEW if new_api else PREMIUM
    open_lookup = OPEN_NEW if new_api else OPEN
    if product_name in premium_lookup:
        return premium_lookup[product_name]
    elif product_name in open_lookup:
        return open_lookup[product_name]
    else:
        raise ValueError(f"{product_name} is not a valid product name")


import unittest.mock as mock
from os import environ

import pytest
from osdatahub.NamesAPI.names_api import NamesAPI

from tests.data import names_data as data

API_KEY = environ.get("OSDATAHUB_TEST_KEY")

class TestFind:
    @pytest.fixture()
    def names(self):
        names = NamesAPI("test")
        yield names

    @pytest.mark.parametrize(*data.test_find_pass())
    @pytest.mark.usefixtures("names")
    @mock.patch('osdatahub.get')
    def test_find_pass(self, request_mocked, names, text, limit, bounds, bbox_filter, local_type,
                       expected_url, expected_params):
        names.find(text, limit=limit, bounds=bounds, bbox_filter=bbox_filter, local_type=local_type)
        request_mocked.assert_called_with(expected_url,
                                          params=expected_params,
                                          proxies={})

    @pytest.mark.parametrize(*data.test_find_fail())
    @pytest.mark.usefixtures("names")
    def test_find_fail(self, names, text, limit, bounds, bbox_filter, local_type, expected_result):
        with pytest.raises(expected_exception=expected_result):
            names.find(text, limit=limit, bounds=bounds, bbox_filter=bbox_filter, local_type=local_type)
    
    
    def test_unauthorised(self):
        with pytest.raises(Exception) as exc_info:
            names = NamesAPI("not a real key")
            names.find("AB22")


    @pytest.mark.skipif(not API_KEY, reason="Test API key not available")
    @pytest.mark.parametrize(*data.test_find_live())
    def test_find_live(self, text, limit, expected_length, minimum_length):
        # Arrange
        names = NamesAPI(API_KEY)
        
        # Act
        results = names.find(text, limit=limit)["features"]
        
        # Assert
        if minimum_length is not None:
            assert len(results) >= minimum_length
        
        if expected_length is not None:
            assert len(results) == expected_length


class TestNearest:
    @pytest.fixture()
    def names(self):
        names = NamesAPI("test")
        yield names

    @pytest.mark.parametrize(*data.test_nearest_pass())
    @pytest.mark.usefixtures("names")
    @mock.patch('osdatahub.get')
    def test_nearest_pass(self, request_mocked, names, point, radius, local_type, expected_url, expected_params):
        names.nearest(point=point, radius=radius, local_type=local_type)
        request_mocked.assert_called_with(expected_url,
                                          params=expected_params,
                                          proxies={})

    @pytest.mark.parametrize(*data.test_nearest_fail())
    @pytest.mark.usefixtures("names")
    def test_nearest_fail(self, names, point, radius, local_type, expected_result):
        with pytest.raises(expected_exception=expected_result):
            names.nearest(point=point, radius=radius, local_type=local_type)


class TestFormatFq:
    @pytest.fixture()
    def names(self):
        key = environ.get("OS_API_KEY")
        names = NamesAPI(key)
        yield names

    @pytest.mark.parametrize(*data.test_format_fq())
    @pytest.mark.usefixtures("names")
    def test_format_fq_pass(self, names, bbox, local_type, expected_result):
        fq_args = names._NamesAPI__format_fq(bbox_filter=bbox, local_type=local_type)
        assert fq_args == expected_result

    @pytest.mark.parametrize(*data.test_format_fq_errors())
    @pytest.mark.usefixtures("names")
    def test_format_fq_errors(self, names, bbox, local_type, expected_result):
        # Act
        with pytest.raises(expected_result):
            fq_args = names._NamesAPI__format_fq(bbox_filter=bbox, local_type=local_type)
